#ifndef MEMORY_H_INCLUDED
#define MEMORY_H_INCLUDED

#include "typedef.h"
#include "device.h"
#include "ini.h"


// ubN
#define MAXRMB		(22)
#define MAXWMB		(17)

// etO
#define MCRCCHK		0b00000001	// CRC`FbNL
#define MUSEEXRAM	0b00000010	// gRAMg
#define MUSESOL		0b00000100	// m̃J[gbWg


////////////////////////////////////////////////////////////////
// NX`
////////////////////////////////////////////////////////////////
class MEM6;

// ubNNX
class MemBlock {
public:
	enum {
		PAGEBITS = 13,	// 8KB
		PAGEMASK = (1 << PAGEBITS) - 1
	};
	
	typedef IDevice::RFuncPtr RFuncPtr;
	typedef IDevice::WFuncPtr WFuncPtr;

protected:
	char Name[33];				// ubN
	BYTE *PRead;				// |C^(Ǎ)
	BYTE *PWrite;				// |C^()
	RFuncPtr FRead;				// ֐|C^(Ǎ)
	WFuncPtr FWrite;			// ֐|C^()
	IDevice *Inst;				// IuWFNg|C^
	int Wait;					// ANZXEFCg
	bool WPt;					// CgveNgtO
	
public:
	MemBlock();									// RXgN^
	~MemBlock();								// fXgN^
	
	void SetMemory( const char *, BYTE *, int, bool );	// 
	void SetRom   ( const char *, BYTE *, int = -1 );	// ROM
	void SetRam   ( const char *, BYTE *, int = -1 );	// RAM
	void SetFunc  ( const char *, BYTE *, IDevice *, RFuncPtr, WFuncPtr, int = -1 );	// ֐
	void SetWait( int );						// ANZXEFCgݒ
	int GetWait() const;						// ANZXEFCg擾
	void SetProtect( bool );					// CgveNgݒ
	bool GetProtect() const;					// CgveNg擾
	
	const char *GetName() const;				// ubN擾
	
	BYTE Read( WORD, int * = NULL ) const;		// [h
	void Write( WORD, BYTE, int * = NULL ) const;	// Cg
};


class MEM6 : public Device, public IDoko {
protected:
	// ROM\
	struct ROMINFO {
		const char *FileName;	// t@C
		DWORD Crc;				// CRC32
	};
	
	// \
	struct MEMINFO {
		const ROMINFO *Rinf;	// ROMւ̃|C^
		const int Rnum;			// ROM̗vf
		DWORD Size;				// TCY
		BYTE Init;				// f[^
		int Wait;				// ANZXEFCg
	};
	
	// ROMe[u\
	struct ROMINFOTABLE {
		const ROMINFO *System1;
		const ROMINFO *System2;
		const ROMINFO *CGRom1;
		const ROMINFO *CGRom2;
		const ROMINFO *Kanji;
		const ROMINFO *Voice;
		
		ROMINFOTABLE() : System1(NULL), System2(NULL), CGRom1(NULL), CGRom2(NULL),
						 Kanji(NULL), Voice(NULL) {}
	};
	
	// e[u\
	struct MEMINFOTABLE {
		const MEMINFO *EmptRom;
		const MEMINFO *EmptRam;
		const MEMINFO *ExtRom;
		const MEMINFO *IntRam;
		const MEMINFO *ExtRam;
		
		const MEMINFO *System1;
		const MEMINFO *System2;
		const MEMINFO *CGRom1;
		const MEMINFO *CGRom2;
		const MEMINFO *Kanji;
		const MEMINFO *Voice;
		
		MEMINFOTABLE() : EmptRom(NULL), EmptRam(NULL), ExtRom(NULL), IntRam(NULL), ExtRam(NULL),
						 System1(NULL), System2(NULL), CGRom1(NULL), CGRom2(NULL),
						 Kanji(NULL), Voice(NULL) {}
	};
	
	// 
	static const MEMINFO IEMPTROM;
	static const MEMINFO IEMPTRAM;
	static const MEMINFO IEXTROM16;
	static const MEMINFO IEXTRAM16;
	static const MEMINFO IEXTRAM64;
	static const MEMINFO IEXTROMS;
	static const MEMINFO IEXTRAMS;
	
	ROMINFOTABLE RomTable;		// ROMe[u
	MEMINFOTABLE MemTable;		// e[u
	
	bool CGBank;				// CG ROM BANK	true:L false:
	bool UseExtRom;				// gROM		true:L false:
	bool UseExtRam;				// gRAM		true:L false:
	
	BYTE *MainRom;				// BASIC ROM	ALL (64,68̎System ROM1)
	BYTE *SysRom2;				// System ROM2	64,68
	BYTE *ExtRom;				// g ROM		ALL
	BYTE *CGRom1;				// CG ROM1		ALL
	BYTE *CGRom2;				// CG ROM2		62,66
	BYTE *KanjiRom;				//  ROM		62,66
	BYTE *VoiceRom;				//  ROM	62,66
	
	BYTE *IntRam;				//  RAM		ALL
	BYTE *ExtRam;				// O RAM		ALL
	
	MemBlock RomB[MAXRMB];		// ROMubN
	MemBlock RamB[MAXWMB];		// RAMubN
	
	MemBlock *Rm_blk[8];		// [hubN|C^(8KB*8)
	MemBlock *Wm_blk[8];		// CgubN|C^(8KB*8)
	
	char FilePath[PATH_MAX];	// gROMt@CtpX
	int M1Wait;					// M1EFCg
	bool EnableChkCRC;			// CRC`FbN  true:L false:
	
	// for 62,66,64,68 -----------------------------------------------------------------------
	bool cgrom;					// CG ROM Ip true:N60[h false:N60m[h
	bool kj_rom;				// ROM,ROM Ip true:ROM false:ROM
	bool kj_LR;					// ROM EIp true:E false:
	bool cgenable;				// CG ROMANZXtO true:ANZX false:ANZXs
	BYTE cgaden;				// CG ROMAhXCl[u
	BYTE cgaddr;				// CG ROMAhX A13,14,15
	BYTE Rf[3];					// Rg[WX^
	// ---------------------------------------------------------------------------------------
	
	// for 64,68 -----------------------------------------------------------------------------
	MemBlock *Rm_blkSR[8];		// [hubN|C^(8KB*8)	SR[hp
	MemBlock *Wm_blkSR[8];		// CgubN|C^(8KB*8)	SR[hp
	BYTE RfSR[16];				// Rg[WX^			SR[hp
	// ---------------------------------------------------------------------------------------
	
	bool AllocMemory( BYTE **, const MEMINFO *, const char * );	// mۂROMt@CǍ
	virtual bool AllocMemorySpecific( const char * ) = 0;		// SmۂROMt@CǍ(@)
	virtual void SetRamValue() = 0;			// RAM̏lݒ
	virtual bool InitSpecific() = 0;		// (@)
	virtual void SetMemBlockR( BYTE, BYTE ) = 0;	// [h̃ubNw(62,66)
	virtual void SetMemBlockW( BYTE ) = 0;	// Cg̃ubNw(62,66)
	
	// for 62,66,64,68 -----------------------------------------------------------------------
	void SetWait( BYTE );					// ANZXEFCgݒ
	BYTE GetWait() const;					// ANZXEFCg擾
	void SetCGrom( BYTE );					// CG ROM AhXݒ(62,66)
	void SelectCGrom( int );				// CG ROM I(62,66)
	void SelectKanjiRom( BYTE );			// ROM  ROM ؂ւ(62,66)
	// ---------------------------------------------------------------------------------------
	
	// m̃J[gbW --------------------------------------------------------------------
	bool UseSol;					// true:L false:
	bool Sol60Mode;					// @[h@true:L false:
	BYTE SolBank[8];				// oNWX^
	int SolBankSet;					// ROMoNZbg
	
	void SetSolBank( BYTE, BYTE );			// oNWX^ݒ
	
	// ubNp֐ ------------------------------------------------------------------
	BYTE SolReadEx( BYTE *, WORD );			// m̃J[gbWǍ(gROM̈)
	BYTE SolMemRead( BYTE *, WORD );		// m̃J[gbWROM/RAMǍ
	void SolMemWrite( BYTE *, WORD, BYTE );	// m̃J[gbWRAM
	BYTE SolSccRead( BYTE *, WORD );		// m̃J[gbWSCCǍ
	void SolSccWrite( BYTE *, WORD, BYTE );	// m̃J[gbWSCC
	// ---------------------------------------------------------------------------------------
	
	// I/OANZX֐ -----------------------------------------------------------------------
	virtual void Out06H( int, BYTE );
	void Out3xH( int, BYTE );
	void Out7FH( int, BYTE );
	
	// for 62,66,64,68 -----------------------------------------------------------------------
	void OutC1H( int, BYTE );
	void OutC2H( int, BYTE );
	void OutC3H( int, BYTE );
	void OutF0H( int, BYTE );
	void OutF1H( int, BYTE );
	void OutF2H( int, BYTE );
	void OutF3H( int, BYTE );
	void OutF8H( int, BYTE );
	BYTE InF0H( int );
	BYTE InF1H( int );
	BYTE InF2H( int );
	BYTE InF3H( int );
	// ---------------------------------------------------------------------------------------
	
public:
	MEM6( VM6 *, const ID& );				// RXgN^
	virtual ~MEM6();						// fXgN^
	
	bool AllocAllMemory( const char *, BYTE );	// SmۂROMt@CǍ
	bool Init();							// 
	virtual void Patch();					// pb`
	virtual void Reset();					// Zbg
	
	BYTE Fetch( WORD, int * = NULL ) const;	// tFb`(M1)
	BYTE Read( WORD, int * = NULL ) const;	// [h
	void Write( WORD, BYTE, int * = NULL ) const;	// Cg
	
	bool MountExtRom( const char * );		// gROM }Eg
	void UnmountExtRom();					// gROM A}Eg
	const char *GetFile() const;			// gROMt@CpX擾
	
	// 8255o͊֘A֐
	virtual void SetCGBank( bool );			// CG ROM BANK ؂ւ
	
	// ڃANZX֐
	BYTE ReadMainRom( WORD ) const;
	BYTE ReadIntRam ( WORD ) const;
	virtual BYTE ReadExtRom ( WORD ) const;
	virtual BYTE ReadExtRam ( WORD ) const;
	virtual BYTE ReadCGrom1 ( WORD ) const;
	virtual BYTE ReadCGrom2 ( WORD ) const;
	virtual BYTE ReadCGrom3 ( WORD ) const;
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	const char *GetReadMemBlk( int ) const ;
	const char *GetWriteMemBlk( int ) const ;
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	
	// foCXID
	enum IDOut{ out06H=0, out3xH, out7FH };
	enum IDIn {};
	
	// ------------------------------------------
	bool DokoSave( cIni * );	// ǂłSAVE
	bool DokoLoad( cIni * );	// ǂłLOAD
	// ------------------------------------------
};


class MEM60 : public MEM6 {
private:
	// ROM
	static const ROMINFO AMAINROM[];
	static const ROMINFO ACGROM1[];
	
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM0;
	static const MEMINFO ICGROM1;
	static const MEMINFO IINTRAM;
	
	bool AllocMemorySpecific( const char * );	// SmۂROMt@CǍ(@)
	void SetRamValue();						// RAM̏lݒ
	bool InitSpecific();					// (@)
	void SetMemBlockR( BYTE, BYTE );		// [h̃ubNw
	void SetMemBlockW( BYTE );				// Cg̃ubNw
	
	// ubNp֐ ------------------------------------------------------------------
	BYTE CGromRead( BYTE *, WORD );			// PC-6001 CGROMǍ
	// ---------------------------------------------------------------------------------------
	
	// foCX`
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
	// I/OANZX֐ -----------------------------------------------------------------------
	void Out06H( int, BYTE );
	
public:
	MEM60( VM6 *, const ID& );				// RXgN^
	~MEM60();								// fXgN^
	
	// foCXID
	enum IDOut{ out06H=0, out3xH, out7FH, outF0H, outF2H };
	enum IDIn {};
};


class MEM61 : public MEM60 {
private:
	// ROM
	static const ROMINFO AMAINROM[];
	static const ROMINFO ACGROM1[];
	
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM0;
	static const MEMINFO ICGROM1;
	static const MEMINFO IINTRAM;
	
public:
	MEM61( VM6 *, const ID& );				// RXgN^
	~MEM61();								// fXgN^
};


class MEM62 : public MEM6 {
protected:
	// ROM
	static const ROMINFO AMAINROM[];
	static const ROMINFO ACGROM1[];
	static const ROMINFO ACGROM2[];
	static const ROMINFO AKANJI[];
	static const ROMINFO AVOICE[];
	
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IINTRAM;
	
	virtual bool AllocMemorySpecific( const char * );	// SmۂROMt@CǍ(@)
	virtual void SetRamValue();				// RAM̏lݒ
	virtual bool InitSpecific();			// (@)
	virtual void SetMemBlockR( BYTE, BYTE );// [h̃ubNw
	void SetMemBlockW( BYTE );				// Cg̃ubNw
	
	// ubNp֐ ------------------------------------------------------------------
	void IERamWrite( BYTE *, WORD, BYTE );	// PC-6001mk2ȍ~ /ORAM
	// ---------------------------------------------------------------------------------------
	
	// foCX`
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
public:
	MEM62( VM6 *, const ID& );				// RXgN^
	virtual ~MEM62();						// fXgN^
	
	// 8255o͊֘A֐
	void SetCGBank( bool );					// CG ROM BANK ؂ւ
	
	// ڃANZX֐
	virtual BYTE ReadCGrom2( WORD ) const;
	virtual BYTE ReadKanjiRom( WORD ) const;
	virtual BYTE ReadVoiceRom( WORD ) const;
	
	// foCXID
	enum IDOut{ out06H=0, out3xH, out7FH, outC1H, outC2H, outC3H, outF0H,  outF1H, outF2H, outF3H, outF8H };
	enum IDIn {                                                    inF0H=0, inF1H,  inF2H,  inF3H         };
};


class MEM66 : public MEM62 {
private:
	// ROM
	static const ROMINFO AMAINROM[];
	static const ROMINFO ACGROM1[];
	static const ROMINFO ACGROM2[];
	static const ROMINFO AKANJI[];
	static const ROMINFO AVOICE[];
	
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IINTRAM;
	
public:
	MEM66( VM6 *, const ID& );				// RXgN^
	~MEM66();								// fXgN^
};


class MEM64 : public MEM62 {
protected:
	// ROM
	static const ROMINFO ASYSROM1[];
	static const ROMINFO ASYSROM2[];
	static const ROMINFO ACGROM[];
	
	// 
	static const MEMINFO ISYSROM1;
	static const MEMINFO ISYSROM2;
	static const MEMINFO ICGROM;
	static const MEMINFO IINTRAM;
	
	bool AllocMemorySpecific( const char * );	// SmۂROMt@CǍ(@)
	virtual void SetRamValue();				// RAM̏lݒ
	bool InitSpecific();					// (@)
	void SetMemBlockR( BYTE, BYTE );		// [h̃ubNw
	void SetMemBlockSR( BYTE, BYTE );		// [h/Cg̃ubNw(64,68)
	
	// foCX`
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
	// I/OANZX֐ -----------------------------------------------------------------------
	void Out6xH( int, BYTE );
	void OutC8H( int, BYTE );
	
	BYTE In6xH( int );
	virtual BYTE InB2H( int );
	
public:
	MEM64( VM6 *, const ID& );				// RXgN^
	virtual ~MEM64();						// fXgN^
	
	void Patch();							// ** pb`
	void Reset();							// ** Zbg
	
	// ڃANZX֐
	BYTE ReadCGrom1( WORD ) const;
	BYTE ReadCGrom2( WORD ) const;
	BYTE ReadCGrom3( WORD ) const;
	BYTE ReadKanjiRom( WORD ) const;
	BYTE ReadVoiceRom( WORD ) const;
	
	// foCXID
	enum IDOut{ out06H=0, out3xH, out7FH, out6xH,   outC1H, outC2H, outC3H, outF0H, outF1H, outF2H, outF3H, outF8H };
	enum IDIn {                            in6xH=0,                          inF0H,  inF1H,  inF2H,  inF3H,
				 inB2H };
	
	// ------------------------------------------
	bool DokoSave( cIni * );	// ǂłSAVE
	bool DokoLoad( cIni * );	// ǂłLOAD
	// ------------------------------------------
};


class MEM68 : public MEM64 {
protected:
	void SetRamValue();						// RAM̏lݒ
	
	BYTE InB2H( int );
	
public:
	MEM68( VM6 *, const ID& );				// RXgN^
	virtual ~MEM68();						// fXgN^
};

#endif	// MEMORY_H_INCLUDED
